from bergen.types.node.ports.base import BasePort


class InputsMeta(type):

  def __new__(mcls, name, bases, attrs):
      return super(InputsMeta, mcls).__new__(mcls, name, bases, attrs)


  @property
  def serialized(self):
      return self._serialized

  def __init__(self, name, bases, attrs):
      super(InputsMeta, self).__init__(name, bases, attrs)
      if attrs["__qualname__"] != "Inputs":
          self._serialized = []

          for key, value in attrs.items():
              if isinstance(value, BasePort):
                self._serialized.append(value(key).serialize())
                 




class Inputs(metaclass=InputsMeta):
  pass

class Outputs(metaclass=InputsMeta):
  pass



from enum import Enum


class GrantType(str, Enum):
    IMPLICIT = "IMPLICIT"
    PASSWORD = "PASSWORD"
    BACKEND = "BACKEND"



class ClientType(str, Enum):
    EXTERNAL = "EXTERNAL"
    INTERNAL = "INTERNAL"
    USER = "USER"


class DataPointType(str, Enum):
    GRAPHQL = "GRAPHQL"
    REST = "REST"



class TYPENAMES(str, Enum):
    MODELPORTTYPE = "ModelPortType"




from abc import ABC, abstractmethod
from asyncio.events import AbstractEventLoop
from typing import Union
from bergen.utils import ExpansionError, expandInputs, shrinkOutputs
from bergen.messages.assignation import AssignationMessage
from bergen.schema import Template
from bergen.constants import ACCEPT_GQL, OFFER_GQL, SERVE_GQL
import logging
import namegenerator
import asyncio
import websockets
from bergen.models import Node, Pod
import inspect
import sys
from aiostream import stream
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from functools import partial
logger = logging.getLogger()

import json


class BaseHelper(ABC):

    def __init__(self, peasent) -> None:
        self.peasent = peasent
        pass

    @abstractmethod
    async def pass_yield(self, message, value):
        pass

    @abstractmethod
    async def pass_progress(self, message, value, percentage=None):
        pass

    @abstractmethod
    async def pass_result(self, message, value):
        pass

    @abstractmethod
    async def pass_exception(self, message, exception):
        pass


class AssignationHelper():

    def __init__(self, peasent_helper: BaseHelper, message: AssignationMessage, loop: AbstractEventLoop = None) -> None:
        self.peasent_helper = peasent_helper
        self.message = message
        self.loop = loop
        pass

    @abstractmethod
    def progress(self, value, percentage=None):
        pass



class ThreadedAssignationHelper(AssignationHelper):

    def progress(self, value, percentage=None):
        logger.info(f'{percentage if percentage else "--"} : {value}')
        if self.loop.is_running():
            future = asyncio.run_coroutine_threadsafe(self.peasent_helper.pass_progress(self.message, value, percentage=percentage), self.loop)
            return future.result()
        else:
            self.loop.run_until_complete(self.peasent_helper.pass_progress(self.message, value, percentage=percentage))


class AsyncAssignationHelper(AssignationHelper):

    async def progress(self, value, percentage=None):
        logger.info(f'{percentage if percentage else "--"} : {value}')
        await self.peasent_helper.pass_progress(self.message, value, percentage=percentage)



async def shrink(node_template_pod, outputs):
    kwargs = {}
    if isinstance(node_template_pod, Node):
        kwargs = await shrinkOutputs(node=node_template_pod, outputs=outputs)

    if isinstance(node_template_pod, Template):
        kwargs = await shrinkOutputs(node=node_template_pod.node, outputs=outputs)

    if isinstance(node_template_pod, Pod):
        kwargs = await shrinkOutputs(node=node_template_pod.template.node, outputs=outputs)


    return kwargs



async def expand(node_template_pod, inputs):
    kwargs = {}
    if isinstance(node_template_pod, Node):
        kwargs = await expandInputs(node=node_template_pod, inputs=inputs)

    if isinstance(node_template_pod, Template):
        kwargs = await expandInputs(node=node_template_pod.node, inputs=inputs)

    if isinstance(node_template_pod, Pod):
        kwargs = await expandInputs(node=node_template_pod.template.node, inputs=inputs) 
    return kwargs



threadhelper = None
try: 
    threadhelper = asyncio.to_thread
except:
    logger.warn("Threading does not work below Python 3.9")

    
class BasePeasent(ABC):
    ''' Is a mixin for Our Bergen '''
    helperClass = None


    def __init__(self, *args, loop = None, name = None, raise_exceptions_local=False, **kwargs) -> None:
        
        
        self.unique_name = name or namegenerator.gen()
        self.loop = loop or asyncio.get_event_loop()
        self.pods = {}


        self.raise_exceptions_local = raise_exceptions_local


        self.nodes_to_register = [] # for Nodes to register
        self.template_function_set = [] # for Templates to register
        self.pod_function_set = [] # For Pods to register

        self.peasent_helper: BaseHelper = self.helperClass(self)

        self.threadpool = ThreadPoolExecutor(5)
        self.processpool = ProcessPoolExecutor(5)

        self.podid_function_map = {}
        self.podid_pod_map = {}
        super().__init__(*args, loop=loop, **kwargs)


    def register(self, node_template_pod: Union[Node, Template], **kwargs):

        threaded = kwargs.get("threaded", False)
        expanded = kwargs.get("threaded", True)


        def real_decorator(function):
            is_coroutine = inspect.iscoroutinefunction(function)
            is_asyncgen = inspect.isasyncgenfunction(function)
            is_function = inspect.isfunction(function)


            async def wrapper(message: AssignationMessage):
                
                
                try:
                    kwargs = await expand(node_template_pod,message.data.inputs)

                    if is_coroutine:
                        assignation_helper = AsyncAssignationHelper(self.peasent_helper, message, loop=self.loop)
                        result = await function(assignation_helper, **kwargs)

                        outputs = await shrink(node_template_pod, result)
                        await self.peasent_helper.pass_result(message, outputs)

                        
                    elif is_asyncgen:
                        assignation_helper = AsyncAssignationHelper(self.peasent_helper, message, loop=self.loop)
                        yieldstream = stream.iterate(function(assignation_helper, **kwargs))
                        lastresult = None
                        async with yieldstream.stream() as streamer:
                            async for result in streamer:
                                lastresult = await shrink(node_template_pod, result)
                                await self.peasent_helper.pass_yield(message, lastresult)

                        await self.peasent_helper.pass_result(message, lastresult)

                    elif is_function:  
                        assignation_helper = ThreadedAssignationHelper(self.peasent_helper, message, loop=self.loop)
                        result = await self.loop.run_in_executor(self.threadpool, partial(function,**kwargs), assignation_helper)

                        outputs = await shrink(node_template_pod, result)
                        await self.peasent_helper.pass_result(message, outputs)

                    else:
                        raise NotImplementedError("We do not allow for non async functions right now")

                except Exception as e:
                    await self.peasent_helper.pass_exception(message, e)
                    logger.error(e)
                    if self.raise_exceptions_local: raise e
                
                
                
            
            wrapper.__name__ = function.__name__

            if isinstance(node_template_pod, Node):
                logger.info(f"Registering Template for {node_template_pod.name}")
                self.nodes_to_register.append({
                    "node": node_template_pod,
                    "params": kwargs,
                    "function": wrapper

                })

            if isinstance(node_template_pod, Template):
                logger.info(f"Already existing Template that we are providing for {node_template_pod.id}")
                self.template_function_set.append((node_template_pod,wrapper))


            if isinstance(node_template_pod, Pod):
                logger.info(f"Already existing Template that we are providing for {node_template_pod.id}")
                self.pod_function_set.append((node_template_pod,wrapper))



            return wrapper


        return real_decorator



    async def setup_and_run(self):

        logger.info("Offering Our Services")
        self.peasent = await SERVE_GQL.run_async(ward=self.main_ward, variables = {"name":self.unique_name})
        logger.info("Our Offer was Accesspted")


        async def offer(node: Node, params: dict, function):
            peasent_template = await OFFER_GQL.run_async(ward=self.main_ward, variables= {"node": node.id, "params": params, "peasent": self.peasent.id})
            return (peasent_template, function)

        logger.info("We are registering our functions as potential Templates")
        template_function_set = await asyncio.gather(*[offer(**values) for values in self.nodes_to_register])

        # We can now combing reigstered templates and already registered templates
        all_templates_function_set =  template_function_set + self.template_function_set

        logger.info("We are our Pods")
        async def accept(template: Template, function):
            peasent_pod = await ACCEPT_GQL.run_async(ward=self.main_ward, variables= {"template": template.id, "peasent": self.peasent.id})
            return (peasent_pod, function)


        pod_function_set = await asyncio.gather(*[accept(*values) for values in all_templates_function_set])
       
        # We can now combing potential templates pods and already potential pods
        all_pod_function_set = pod_function_set + self.pod_function_set

        self.podid_function_map = {value[0].id: value[1] for value in all_pod_function_set}
        self.podid_pod_map = {value[0].id: value[0] for value in all_pod_function_set}


        logger.info(f"Following functions have been allowed! {[value[1].__name__ for value in all_pod_function_set]}")
        for value in all_pod_function_set:
            logger.info(f"Requesting to host {value[1].__name__} as Pod {value[0].id}" )


        logger.debug("Configuring")
        await self.configure()



    @abstractmethod
    async def configure(self) -> str:
        raise NotImplementedError("Please overwrite")

    async def run_async(self):
        await self.setup_and_run()

    def run(self):
        if self.loop.is_running():
            logger.error("Cannot do this, please await run_asnyc()")
        else:
            task = self.loop.create_task(self.setup_and_run())

            async def _async(task):
                return await asyncio.wait([task])

            self.loop.run_until_complete(_async(task))

        # we enter a never-ending loop that waits for data
        # and runs callbacks whenever necessary.
        




import asyncio
from bergen.messages.assignation import AssignationMessage
from bergen.schema import Template
from bergen.peasent.base import BasePeasent
import json
import aiormq
from aiostream import stream
    


class PikaPeasent(BasePeasent):
    """ BROKEN AS FOR NOW """

    async def on_message(self, message:aiormq.types.DeliveredMessage):

        assignation = AssignationMessage.from_message(message=message)

        message_body = json.loads(message.body.decode())

        assert "meta" in message_body, "No meta provided, check Arnheim Schema"
        assert "data" in message_body, "No meta provided, check Arnheim Schema"


        if "progress" in message_body["meta"]:
            print("Updating Progress")
            await message.channel.basic_publish(
            "Started".encode(), routing_key=message_body["meta"]["progress"],
            properties=aiormq.spec.Basic.Properties(
                correlation_id=message.header.properties.correlation_id
            ),

        )

        #xs = stream.count(interval=10)

        #async with xs.stream() as streamer:

          #async for t in streamer:
        
        def worker(inputs):
            return {"rep": 1}

        result = worker(assignation.data.inputs)

        assignation.data.outputs = result

        await message.channel.basic_publish(
            assignation.to_message(), routing_key=message.header.properties.reply_to,
            properties=aiormq.spec.Basic.Properties(
                correlation_id=message.header.properties.correlation_id
            ),

        )
        print(f"Yielding {assignation.data.id}")

        await message.channel.basic_ack(message.delivery.delivery_tag)
        print('Request complete')


    async def configure(self):
        # Perform connection
        connection = await aiormq.connect(f"amqp://{self.token}:guest@localhost/")
        channel = await connection.channel()

        # Declaring queue
        pod_assignment_queue = await channel.queue_declare('pod_two')

        # Start listening the queue with name 'hello'
        await channel.basic_consume(pod_assignment_queue.queue, self.on_message)

        return pod_assignment_queue.queue









    


import asyncio
from bergen.wards.graphql.aiohttp import AIOHttpGraphQLWard
from bergen.enums import ClientType
from bergen.logging import setLogging
from bergen.auths.base import BaseAuthBackend
from bergen.wards.base import BaseWard
from bergen.postmans.base import BasePostman
import logging
from threading import Thread


logger = logging.getLogger(__name__)
import os

def start_background_loop(loop: asyncio.AbstractEventLoop) -> None:
    asyncio.set_event_loop(loop)

class BaseBergen:


    def __init__(self, auth: BaseAuthBackend= None, host: str = None, port: int = None, ssl=False, auto_negotiate=True, bind=True, log=logging.INFO, local=None, loop=None,  client_type: ClientType = None, jupyter=False, force_sync = False, **kwargs) -> None:
        

        if jupyter:
            setLogging(logging.ERROR)
        else:
            print(r"     _               _          _              ____ _ _            _    ")   
            print(r"    / \   _ __ _ __ | |__   ___(_)_ __ ___    / ___| (_) ___ _ __ | |_  ")
            print(r"   / _ \ | '__| '_ \| '_ \ / _ \ | '_ ` _ \  | |   | | |/ _ \ '_ \| __| ")
            print(r"  / ___ \| |  | | | | | | |  __/ | | | | | | | |___| | |  __/ | | | |_  ")
            print(r" /_/   \_\_|  |_| |_|_| |_|\___|_|_| |_| |_|  \____|_|_|\___|_| |_|\__| ")
            print(r"")
            setLogging(log)


        self.running_in_sync = force_sync
        

        
        self.loop = asyncio.get_event_loop()
        if self.loop.is_running():
            if self.running_in_sync:
                import nest_asyncio
                nest_asyncio.apply(self.loop)
            self.loop_is_running = True
        else:
            self.loop_is_running = False

        if bind: 
            # We only import this here for typehints
            from bergen.registries.arnheim import set_current_arnheim
            set_current_arnheim(self)

        

        self.local = local if local is not None else os.getenv("ARNHEIM_LOCAL") == "1" 

        if self.local:
            logger.info("Running in Local Mode")

        self.client_type = client_type


        self.auth = auth
        self.token = self.auth.getToken()
        logger.info(" Auhorized!!!!!")

        self.host = host
        self.port = port
        self.protocol = "https" if ssl else "http"

        self._transcript = None
        self.identifierDataPointMap = {}
        self.identifierWardMap: dict[str, BaseWard] = {}


        if auto_negotiate == True:
            if self.loop.is_running() and not self.running_in_sync: 
                pass
            else:
               self.negotiate()


        super().__init__() 

    def getLoopAndContext(self):
        return self.loop, self.running_in_sync    


    @property
    def transcript(self):
        assert self._transcript is not None, "We have to negotiate first with our"
        return self._transcript

    def getExtensionSettings(self, extension):
        assert extension in self.transcript.extensions, f"Arnheim seems to have no idea about this Extension {extension}"
        return self.transcript.extensions[extension]

    def getWardForIdentifier(self, identifier):
        if self._transcript is None:
            if self.running_in_sync:
                raise Exception("Not negotiated Error: Please negotiate first or set auto_negotiate=True")
            else:
                raise Exception("You are running in event Loop: Please await self.negotiate_async first or run with 'async with Bergen(....)'")

        if identifier in self.identifierWardMap:
            return self.identifierWardMap[identifier]
        else:
            return self.main_ward


    def getPostmanFromSettings(self, transcript):
        settings = transcript.postman

        if settings.type == "pika":
            try:
                from bergen.postmans.pika import PikaPostman
                postman = PikaPostman(**settings.kwargs, loop=self.loop)
            except ImportError as e:
                logger.error("You cannot use the Pika Postman without installing aio_pika")
                raise e

        elif settings.type == "websocket":
            try:
                from bergen.postmans.websocket import WebsocketPostman
                postman = WebsocketPostman(**settings.kwargs, loop=self.loop)
            except ImportError as e:
                logger.error("You cannot use the Websocket Postman without installing websockets")
                raise e

        else:
            raise Exception(f"Postman couldn't be configured. No Postman for type {settings.type}")

        return postman

    
    async def negotiate_async(self, client_type=None):
        from bergen.constants import NEGOTIATION_GQL
        from bergen.registries.datapoint import get_datapoint_registry

        # Instantiate our Main Ward, this is only for Nodes and Pods
        self.main_ward = AIOHttpGraphQLWard(host=self.host, port=self.port, protocol=self.protocol, token=self.token, loop=self.loop)
        await self.main_ward.configure()

        # We resort escalating to the different client Type protocols
        clientType = client_type or self.client_type or self.auth.getClientType()
        self._transcript = await NEGOTIATION_GQL.run_async(ward=self.main_ward, variables={"clientType": clientType, "local": self.local})
        

        #Lets create our different Wards 
        
        assert self._transcript.models is not None, "We apparently didnt't get any points"
        
        datapoint_registry = get_datapoint_registry()


        self.identifierDataPointMap = {model.identifier.lower(): model.point for model in self._transcript.models}
        self.identifierWardMap = {model.identifier.lower(): datapoint_registry.createWardForDatapoint(model.point, self) for model in self._transcript.models}

        logger.info("Succesfully registered Datapoints") 
        await datapoint_registry.configureWards()
        logger.info("Succesfully connected to Datapoints") 


        self.postman = self.getPostmanFromSettings(self._transcript)
        await self.postman.configure()

    async def disconnect_async(self, client_type=None):
        await self.postman.disconnect()

    def negotiate(self, client_type = None):
        self.loop.run_until_complete(self.negotiate_async())
        
    def getWard(self) -> BaseWard:
        return self.main_ward

    def getPostman(self) -> BasePostman:
        return self.postman

    def _repr_html_(self):
        if not self._transcript: return """Unconnected Client"""
        return f"""
            <p> Arnheim Client <p>
            <table>
                <tr>
                    <td> Connected to </td> <td> {self.main_ward.name} </td>
                </tr>
            </table>

        """

    async def __aenter__(self):
        await self.negotiate_async()
        return self


    async def __aexit__(self,*args, **kwargs):
        await self.disconnect_async()







from bergen.messages.assignation import AssignationMessage
from bergen.messages.exceptions.base import ExceptionMessage
from bergen.messages.types import ASSIGNATION, EXCEPTION
from bergen.messages.base import MessageModel
from bergen.messages.assignation_request import AssignationRequestMessage
from bergen.postmans.base import BasePostman
from aiostream import stream
import aiormq
import json
import uuid
import logging
import asyncio

logger = logging.getLogger(__name__)






class PikaPostman(BasePostman):
    type = "pika"

    def __init__(self, port= None, protocol = None, host= None, auth= None, **kwargs) -> None:
        self.token = auth["token"]
        self.port = port
        self.protocol = protocol
        self.host = host
        self.connection = None      # type: aiormq.Connection
        self.channel = None         # type: aiormq.Channel
        self.callback_queue = ''
        self.futures = {}

        self.configured = False

        self.assign_routing = "assignation_request"
        super().__init__(**kwargs)

    async def configure(self):

        self.loop = asyncio.get_event_loop()
        self.connection = await aiormq.connect(f"amqp://{self.token}:guest@{self.host}/")

        self.channel = await self.connection.channel()
        declare_ok = await self.channel.queue_declare(
            exclusive=True, auto_delete=True
        )
        declare_progress = await self.channel.queue_declare(
            exclusive=True, auto_delete=True
        )

        await self.channel.basic_consume(declare_ok.queue, self.on_response)
        await self.channel.basic_consume(declare_progress.queue, self.on_progress)

        self.callback_queue = declare_ok.queue
        self.progress_queue = declare_progress.queue

    async def on_response(self, message: aiormq.types.DeliveredMessage):
        print("Received somethign here")
        correlationid = message.header.properties.correlation_id
        if correlationid not in self.futures: 
            print("Received Uncorrelated")
            return # TODO: Handle this? We coudlnt care less

        parsed_message = MessageModel.from_message(message=message)
        future = self.futures.pop(message.header.properties.correlation_id)
        if parsed_message.meta.type == EXCEPTION:
            parsed_exception = ExceptionMessage.from_message(message=message)
            future.set_exception(Exception(parsed_exception.data.message))
        if parsed_message.meta.type == ASSIGNATION:
            parsed_assignation = AssignationMessage.from_message(message=message)
            future.set_result(parsed_assignation.data.outputs)
        else:
            raise Exception("Received something weird", message.body.decode() )

    async def on_progress(self, message: aiormq.types.DeliveredMessage):
        logger.info(message.body.decode())


    async def assign(self, node, inputs, params, **extensions):
        correlation_id = str(uuid.uuid4()) # Where should we do this?
        future = self.loop.create_future()
        self.futures[correlation_id] = future
       

        request = AssignationRequestMessage(data={
                                                "node":node.id, 
                                                "inputs": inputs or {}, 
                                                "params": dict(params or {}),
                                                "reference": correlation_id,
                                                "callback": self.callback_queue,
                                                "progress": self.progress_queue,
                                            },
                                            meta={
                                                "auth": {
                                                    "token": self.token
                                                },
                                                "extensions": extensions
                                            })
            
        await self.channel.basic_publish(
            str(json.dumps(request.dict())).encode(), routing_key=self.assign_routing,
            properties=aiormq.spec.Basic.Properties(
                content_type='application/json',
                correlation_id=correlation_id,
                reply_to=self.callback_queue, # If we have a malformed Message
            )
        )

        return await future


    async def delay(self, node, inputs, params, **extensions):
        
        reference = str(uuid.uuid4())
        
        request = AssignationRequestMessage(data={
                                                "node":node.id, 
                                                "inputs": inputs, 
                                                "params": dict(params or {}),
                                                "reference": reference,
                                                # We omit sending a callback
                                            },
                                            meta={
                                                "auth": {
                                                    "token": self.token
                                                },
                                                "extensions": extensions
                                            })

        await self.channel.basic_publish(
            str(json.dumps(request.dict())).encode(), routing_key=self.assign_routing,
            properties=aiormq.spec.Basic.Properties(
                content_type='application/json',
                correlation_id=reference,
                reply_to=self.callback_queue, #TODO: There might be not a need for this?
            )
        )

        return reference



from bergen.query import DelayedGQL


POD_QUERY = DelayedGQL("""
query Pod($id: ID){
  pod(id: $id){
    id
    status
    name
    unique
    channel
    template {
      id
      provider {
        name
      }
      node {
        id
        name
        image
        inputs {
          __typename
          key
          required
          ... on ModelPortType {
            identifier
          }
        }
        outputs {
          __typename
          key
          required
          ... on ModelPortType {
            identifier
          }
        }
      }
    }
  }
}
""")



from bergen.extenders.base import BaseExtender
import logging

logger = logging.getLogger(__name__)

class PortExtender(BaseExtender):


    def provide(template = None, provider = None, **kwargs):
        logger.info("Starting provision")



    def __call__(inputs: dict, provider="vard"):
        logger.info("Called with inputs")

        logger.info("Serializing Inputs")


    def _repr_html_(self):
        string = f"{self.TYPENAME} with {self.key}"

        return string




from bergen.schema import PodStatus
import logging


logger = logging.getLogger(__name__)
class HostedPod():

    def __init__(self, node, params, enter_when: PodStatus = PodStatus.ACTIVE, **kwargs) -> None:
        self.node = node
        self.assigned_pod = None
        self.provision_kwargs = kwargs
        self.provision_params = params
        self.enter_when = enter_when
        pass



    async def __aenter__(self) -> "Pod":
        self.assigned_pod = await self.node.provide_async(self.provision_params, **self.provision_kwargs, return_when=self.enter_when)
        return self.assigned_pod

    async def __aexit__(self, exc_type, exc, tb):
        logger.info('Exiting Rabbit')
        await self.assigned_pod.unprovide()


import logging

def setLogging(log_level):
    LOGFORMAT = "  s%(levelname)-8s%(reset)s | %(log_color)s%(message)s%(reset)s"
    stream = logging.StreamHandler()
    logging.root.setLevel(log_level)
    stream.setLevel(log_level)

    try:
        from colorlog import ColoredFormatter
        LOGFORMAT = "  %(log_color)s%(levelname)-8s%(reset)s | %(log_color)s%(message)s%(reset)s"
        formatter = ColoredFormatter(LOGFORMAT)
        stream.setFormatter(formatter)
    except:
        pass

    logger = logging.getLogger()
    logger.setLevel(log_level)
    logger.addHandler(stream)



from abc import ABC
from bergen.constants import VOLUNTEER_GQL
import logging
from typing import Callable
from bergen.schema import Volunteer, Pod, Node


logger = logging.getLogger(__name__)


class HostHelper(ABC):


    def __init__(self, arnheim) -> None:
        self.arnheim = arnheim
        self.identifierFunctionMap: dict[str, Callable] = {}
        self.identifierVolunteerMap: dict[str, Volunteer] = {}
        self.identifierPodMap: dict[str, Pod] = {}
        self.podIDIdentifierMap: dict[int, str] = {}

    def addVolunteer(self, volunteer: Volunteer):
        self.identifierVolunteerMap[volunteer.identifier] = volunteer
        logger.info(f"Successfully added Volunteer for {volunteer.node.name}: VolunteerID: {volunteer.id}")

    def addPod(self, pod: Pod):
        self.identifierPodMap[pod.volunteer.identifier] = pod
        self.podIDIdentifierMap[pod.id] = pod.volunteer.identifier
        logger.info(f"Successfully added Pod for {pod.id}")


    def addVolunteerFunction(self, volunteer: Volunteer, callable: Callable):
        self.identifierFunctionMap[volunteer.identifier] = callable


    def getFunctionForPod(self, pod: Pod):
        return self.identifierFunctionMap[self.getIdentifierForPod(pod)]

    def getVolunteers(self):
        return [volunteer for key, volunteer in self.identifierVolunteerMap.items()]
        
    def volunteer(self, node: Node, version="standard"):
        # Implement getting node
        volunteer = VOLUNTEER_GQL.run(ward=self.arnheim.main_ward, variables = {"nodeid": node.id, "name": self.arnheim.name, "version": version})
        self.addVolunteer(volunteer)
        return volunteer


    def getIdentifierForPod(self, pod: Pod):
        return self.podIDIdentifierMap[pod.id]

    def getPodForIdentifier(self, identifier):
        return self.identifierPodMap[identifier]

    def getVolunteerForPod(self, pod: Pod):
        return self.identifierVolunteerMap[self.getIdentifierForPod(pod)]


from abc import abstractmethod
from abc import ABC
from bergen.query import  TypedGQL
from typing import TypeVar
import namegenerator

T = TypeVar("T")

class WardException(Exception):
    pass




class BaseWard(ABC):

    def __init__(self, loop=None, name = None):
        self.loop = loop
        self.name = name or namegenerator.gen()

    @abstractmethod
    async def configure(self):
        pass

    def run(self, the_query: TypedGQL, variables: dict = {}, **kwargs):
        return self.loop.run_until_complete(self.run_async(the_query, variables=variables, **kwargs))

    @abstractmethod
    def run_async(self, gql: TypedGQL, variables: dict = {}):
        return gql.cls(**{})


from asyncio.tasks import ensure_future
from bergen.messages.provision_request import ProvisionRequestMessage
from bergen.utils import expandOutputs, shrinkInputs
from bergen.messages.assignation import AssignationMessage
from bergen.messages.provision import ProvisionMessage
from bergen.messages.exceptions.base import ExceptionMessage
from bergen.messages.types import ASSIGNATION, EXCEPTION, PROVISION
from bergen.messages.base import MessageModel
from bergen.messages.assignation_request import AssignationRequestMessage
from bergen.postmans.base import BasePostman
from aiostream import stream
import json
import uuid
import logging

import asyncio
try:
    from asyncio import create_task
except ImportError:
    #python 3.6 fix
    create_task = asyncio.ensure_future


import websockets
import sys
from bergen.schema import AssignationStatus
from bergen.models import Pod



logger = logging.getLogger(__name__)


class NodeException(Exception):
    pass



class WebsocketPostman(BasePostman):
    type = "websocket"

    def __init__(self, port= None, protocol = None, host= None, auth= None, **kwargs) -> None:
        self.token = auth["token"]
        self.port = port
        self.protocol = protocol
        self.host = host
        self.connection = None      
        self.channel = None         
        self.callback_queue = ''


        self.futures = {}
        self.progresses = {}




        self.provision_futures = {}
        self.unprovision_futures = {}

        self.configured = False
        self.consumer_task = None
        self.producer_task = None

        self.assign_routing = "assignation_request"
        self.active_stream_queues = {}
        super().__init__(**kwargs)

    async def configure(self):
        self.callback_queue = asyncio.Queue()
        self.progress_queue = asyncio.Queue()
        self.startup_task = create_task(self.startup())


    async def disconnect(self):
        if self.startup_task: self.startup_task.cancel()
        if self.consumer_task: self.consumer_task.cancel()
        if self.producer_task: self.producer_task.cancel()
        print("Bergen disconnected")

    async def startup(self):

        await self.connect_websocket()

        self.consumer_task = create_task(
            self.producer()
        )
        self.producer_task = create_task(
            self.workers()
        )

        done, pending = await asyncio.wait(
            [self.consumer_task, self.producer_task],
            return_when=asyncio.ALL_COMPLETED
        )

        for task in pending:
            task.cancel()


    async def connect_websocket(self):
        uri = f"ws://{self.host}:{self.port}/postman/?token={self.token}"
        try:
            self.connection = await websockets.client.connect(uri)
        except ConnectionRefusedError:
            sys.exit('error: cannot connect to backend')


    async def producer(self):
        logger.info(" [x] Awaiting Reponse from Postman")
        async for message in self.connection:
            await self.callback_queue.put(message)
        # await self.send_queue.put(message)

    async def workers(self):
        while True:
            message = await self.callback_queue.get()
            try:
                parsed_message = MessageModel.from_channels(message=message)
                # Stream Path
            
                if parsed_message.meta.type == EXCEPTION: #Protocol Exception
                    parsed_exception = ExceptionMessage.from_channels(message=message)
                    future = self.futures.pop(parsed_exception.meta.reference)
                    future.set_exception(Exception(parsed_exception.data.message))


                elif parsed_message.meta.type == ASSIGNATION: # Assignation Exception
                    parsed_assignation = AssignationMessage.from_channels(message=message)

                    correlation_id = parsed_assignation.data.reference

                    if correlation_id in self.active_stream_queues: # It is a stream delage to stream
                        await self.active_stream_queues[correlation_id].put(parsed_assignation)
                    
                    if correlation_id in self.futures:

                        if parsed_assignation.data.status == AssignationStatus.PROGRESS:
                            if correlation_id in self.progresses:
                                self.progresses[correlation_id](parsed_assignation.data.statusmessage) # call the function that is the progress function

                        if parsed_assignation.data.status == AssignationStatus.DONE:
                            if correlation_id in self.progresses:
                                self.progresses.pop(correlation_id)
                            future = self.futures.pop(correlation_id)
                            future.set_result(parsed_assignation.data.outputs)

                        if parsed_assignation.data.status == AssignationStatus.ERROR:
                            if correlation_id in self.progresses:
                                self.progresses.pop(correlation_id)
                            future = self.futures.pop(correlation_id)
                            future.set_exception(NodeException(parsed_assignation.data.statusmessage))

                elif parsed_message.meta.type == PROVISION:
                    print(message)
                    parsed_provision = ProvisionMessage.from_channels(message=message)

                    correlation_id = parsed_provision.meta.reference

                    if correlation_id in self.active_stream_queues:
                        await self.active_stream_queues[correlation_id].put(parsed_provision)
                    
                    if correlation_id in self.provision_futures:
                        future = self.provision_futures.pop(correlation_id)
                        future.set_result(parsed_provision.data.pod)

                else:
                    raise Exception("Received something weird", parsed_message )

            except Exception as e:
                logger.error(e)
            self.callback_queue.task_done()

    async def stream(self, node, inputs, params, **extensions):
        logger.info(f"Creating a Stream of Data for {node.id}")
        correlation_id = str(uuid.uuid4())

        assigned_inputs = await shrinkInputs(node=node, inputs=inputs)

        self.active_stream_queues[correlation_id] = asyncio.Queue()

        request = AssignationRequestMessage(data={
                                                "node":node.id, 
                                                "inputs": assigned_inputs, 
                                                "params": dict(params or {}),
                                                "reference": correlation_id,
                                                # We omit sending a callback
                                            },
                                            meta={
                                                "reference": correlation_id,
                                                "auth": {
                                                    "token": self.token
                                                },
                                                "extensions": extensions
                                            })

        await self.send_to_arnheim(request)

        while True:
            messagein_stream = await self.active_stream_queues[correlation_id].get()

            if messagein_stream.data.status == AssignationStatus.YIELD:
                yield await expandOutputs(node, outputs=messagein_stream.data.outputs)

            if messagein_stream.data.status == AssignationStatus.DONE:
                break


    async def stream_progress(self, node, inputs, params, **extensions):
        logger.debug(f"Creating a Stream of Progress for {node.id}")
        correlation_id = str(uuid.uuid4())

        assigned_inputs = await shrinkInputs(node=node, inputs=inputs)

        self.active_stream_queues[correlation_id] = asyncio.Queue()

        request = AssignationRequestMessage(data={
                                                "node":node.id, 
                                                "inputs": assigned_inputs, 
                                                "params": dict(params or {}),
                                                "reference": correlation_id,
                                                # We omit sending a callback
                                            },
                                            meta={
                                                "reference": correlation_id,
                                                "auth": {
                                                    "token": self.token
                                                },
                                                "extensions": {
                                                    "with_progress": True,
                                                    **extensions
                                                }
                                            })

        await self.send_to_arnheim(request)

        while True:
            messagein_stream = await self.active_stream_queues[correlation_id].get()
            if messagein_stream.data.status == AssignationStatus.PROGRESS:
                yield messagein_stream.data.statusmessage

            if messagein_stream.data.status == AssignationStatus.ERROR:
                raise NodeException(messagein_stream.data.statusmessage)

            if messagein_stream.data.status == AssignationStatus.DONE:
                yield messagein_stream.data.outputs
                break




    async def send_to_arnheim(self,request):
        if self.connection:
            await self.connection.send(request.to_channels())
        else:
            raise Exception("No longer connected. Did you use an Async context manager?")
        

    async def assign(self, node_or_pod, inputs, params, **extensions):
        correlation_id = str(uuid.uuid4()) # Where should we do this?
        future = self.loop.create_future()

        self.futures[correlation_id] = future
        
        if "with_progress" in extensions:
            self.progresses[correlation_id] = lambda progress: logger.info(progress) # lets look for progress?


        assigned_inputs = await shrinkInputs(node=node_or_pod, inputs=inputs)

       
        #TODO: Implement assigning to pod directly
        request = AssignationRequestMessage(data={
                                                "node":node_or_pod.id, 
                                                "inputs": assigned_inputs, 
                                                "params": dict(params or {}),
                                                "reference": correlation_id,
                                                "callback": None,
                                                "progress": None,
                                            },
                                            meta={
                                                "reference": correlation_id,
                                                "auth": {
                                                    "token": self.token
                                                },
                                                "extensions": extensions
                                            })


        await self.send_to_arnheim(request)
            
        outputs = await future
        return await expandOutputs(node_or_pod, outputs)


    async def provide(self, node, params, **extensions):
        correlation_id = str(uuid.uuid4()) # Where should we do this?
        future = self.loop.create_future()
        self.provision_futures[correlation_id] = future

        request = ProvisionRequestMessage(data={
                                                "node":node.id, 
                                                "params": dict(params or {}),
                                                "reference": correlation_id,
                                            },
                                            meta={
                                                "reference": correlation_id,
                                                "auth": {
                                                    "token": self.token
                                                },
                                                "extensions": extensions
                                            })


        await self.send_to_arnheim(request)
            

        pod_id = await future
        pod = await Pod.asyncs.get(id=pod_id)
        return pod


    async def unprovide(self, pod):
        correlation_id = str(uuid.uuid4()) # Where should we do this?
        #future = self.loop.create_future()
        #self.unprovision_futures[correlation_id] = future

        print("Unproviding not implememted yet")           

        return False


    async def delay(self, node, inputs, params, **extensions):
        
        reference = str(uuid.uuid4())
        assigned_inputs = await shrinkInputs(node=node, inputs=inputs)
        request = AssignationRequestMessage(data={
                                                "node":node.id, 
                                                "inputs": assigned_inputs, 
                                                "params": dict(params or {}),
                                                "reference": reference,
                                                # We omit sending a callback
                                            },
                                            meta={
                                                "auth": {
                                                    "token": self.token
                                                },
                                                "extensions": extensions
                                            })

        await self.send_to_arnheim(request)

        return reference



from typing import Any
from bergen.schema import AssignationParams, Pod, PodStatus, ProvisionParams
from bergen.registries.arnheim import get_current_arnheim
import asyncio
from bergen.types.model import ArnheimModel
from bergen.extenders.base import BaseExtender
import logging
from aiostream import stream
import uuid
from bergen.extenders.contexts.pod import HostedPod
from tqdm import tqdm
import textwrap

class NodeExtender(BaseExtender):


    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args,**kwargs)
        
        bergen = get_current_arnheim()

        self._postman = bergen.getPostman()
        self._loop, self._force_sync = bergen.getLoopAndContext()


    def provide(self, params: ProvisionParams, **kwargs):
        if self._loop.is_running() and not self._force_sync:
            return self.provide_async(params, **kwargs)
        else:
            result = self._loop.run_until_complete(self.provide_async(params, **kwargs))
            return result

    async def provide_async(self, params: ProvisionParams, **kwargs):
        return await self._postman.provide(self, params, **kwargs)       

    async def assign_async(self, inputs: dict, params: AssignationParams, **kwargs):
        
        return await self._postman.assign(self, inputs, params, **kwargs)

    async def delay_async(self, inputs: dict, params: AssignationParams, **kwargs):
    
        return await self._postman.delay(self, inputs, params, **kwargs)

    def stream(self, inputs: dict, params: AssignationParams = None, **kwargs):

        return stream.iterate(self._postman.stream(self, inputs, params, **kwargs)).stream()


    async def assign_with_progress(self, inputs, params, **kwargs):
        result = None
        with tqdm(total=100) as pbar:
                async with self.stream_progress(inputs, params, **kwargs) as stream:
                        async for item in stream:
                                result = item
                                if isinstance(result, dict): break
                                
                                progress, message = item.split(":")
                                try: 
                                        pbar.n = int(progress)
                                        pbar.refresh()
                                except:
                                        pass
                                pbar.set_postfix_str(textwrap.shorten(message, width=30, placeholder="..."))
                pbar.n = 100
                pbar.refresh()
                pbar.set_postfix_str("Done")
        return result


    def stream_progress(self,  inputs: dict, params: AssignationParams = None, **kwargs):
        return stream.iterate(self._postman.stream_progress(self, inputs, params, **kwargs)).stream()
    
    def delay(self, inputs: dict, params: AssignationParams = None, **kwargs):
        if self._loop.is_running() and not self._force_sync:
            return self.delay_async(inputs, params, **kwargs)
        else:
            result = self._loop.run_until_complete(self.delay_async(inputs, params, **kwargs))
            return result


    def __call__(self, inputs: dict, params: AssignationParams = None, with_progress = False, **kwargs) -> dict:
        """Call this node (can be run both asynchronously and syncrhounsly)

        Args:
            inputs (dict): The inputs for this Node
            params (AssignationParams, optional): [description]. Defaults to None.

        Returns:
            outputs (dict): The ooutputs of this Node
        """
        if self._loop.is_running() and not self._force_sync:
            if with_progress == True:
                return self.assign_with_progress(inputs, params,  with_progress=with_progress, **kwargs)
            return self.assign_async(inputs, params, with_progress=with_progress, **kwargs)

        else:
            if with_progress == True:
                return self._loop.run_until_complete(self.assign_with_progress(inputs, params,  with_progress=with_progress, **kwargs))

            result = self._loop.run_until_complete(self.assign_async(inputs, params,  with_progress=with_progress, **kwargs))
            return result



    def hosted(self, params: ProvisionParams = {}, enter_when = PodStatus.ACTIVE, **kwargs):
        return HostedPod(self, params = {}, enter_when=enter_when, **kwargs)

    def _repr_html_(self):
        string = f"{self.name}</br>"

        for input in self.inputs:
            string += "Inputs </br>"
            string += f"Port: {input._repr_html_()} </br>"

        for output in self.outputs:
            string += "Outputs </br>"
            string += f"Port: {output._repr_html_()} </br>"


        return string


from abc import ABC, abstractmethod
from aiostream import stream


class BasePostman(ABC):
    """ A Postman takes node requests and translates them to Bergen calls, basic implementations are GRAPHQL and PIKA"""

    def __init__(self, requires_configuration=True, loop=None) -> None:
        assert loop is not None, "Please provide a Loop to your Postman, Did you forget to call super.init with **kwargs?"
        self.loop = loop

    @abstractmethod
    async def configure(self):
        pass

    @abstractmethod
    async def stream(inputs: dict, params: dict, **kwargs):
        return NotImplementedError( "Abstract class")


    @abstractmethod
    async def assign(self, node, inputs: dict, params: dict, **kwargs):

        return NotImplementedError("This is abstract")

    @abstractmethod
    async def provide(self, params: dict, **kwargs):

        return NotImplementedError("This is abstract")

    @abstractmethod
    async def unprovide(self):

        return NotImplementedError("This is abstract")

    @abstractmethod
    async def delay(self, inputs: dict, params: dict, **kwargs):

        return NotImplementedError("This is abstract")



